# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the background_2d module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.nddata import CCDData, NDData
from astropy.stats import SigmaClip
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose, assert_equal

from photutils.background import (Background2D, BkgZoomInterpolator,
                                  MeanBackground, MedianBackground,
                                  SExtractorBackground)
from photutils.utils._optional_deps import HAS_MATPLOTLIB


@pytest.fixture
def test_data():
    """
    Create test data for Background2D tests.
    """
    return np.ones((100, 100))


@pytest.fixture
def bkg_rms(test_data):
    """
    Expected background RMS for test data.
    """
    return np.zeros(test_data.shape)


@pytest.fixture
def bkg_mesh():
    """
    Expected background mesh for test data.
    """
    return np.ones((4, 4))


@pytest.fixture
def bkg_rms_mesh():
    """
    Expected background RMS mesh for test data.
    """
    return np.zeros((4, 4))


@pytest.fixture(params=['quantity', 'nddata_with_unit', 'ccddata'])
def nddata_variant(request, test_data):
    """
    Create different variants of input data with units for testing.
    """
    if request.param == 'quantity':
        return test_data << u.ct
    if request.param == 'nddata_with_unit':
        return NDData(test_data, unit=u.ct)
    return CCDData(test_data, unit=u.ct)


@pytest.fixture
def nddata_no_unit(test_data):
    """
    Create NDData without units for testing.
    """
    return NDData(test_data, unit=None)


class TestBackground2D:
    """
    Test the Background2D class.
    """
    @pytest.mark.parametrize('filter_size', [(1, 1), (3, 3)])
    def test_background(self, filter_size, test_data, bkg_rms, bkg_mesh,
                        bkg_rms_mesh):
        """
        Test with different filter sizes.
        """
        bkg = Background2D(test_data, (25, 25), filter_size=filter_size)
        assert_allclose(bkg.background, test_data)
        assert_allclose(bkg.background_rms, bkg_rms)
        assert_allclose(bkg.background_mesh, bkg_mesh)
        assert_allclose(bkg.background_rms_mesh, bkg_rms_mesh)
        assert bkg.background_median == 1.0
        assert bkg.background_rms_median == 0.0
        assert bkg.n_pixels_mesh.shape == (4, 4)
        assert bkg.n_pixels_map.shape == test_data.shape

    @pytest.mark.parametrize('box_size', [(25, 25), (23, 22)])
    @pytest.mark.parametrize('dtype', ['int', 'int32', 'float32'])
    def test_background_dtype(self, box_size, dtype, test_data, bkg_rms):
        """
        Test that the output background and RMS have the same dtype as
        the input data, or are floating point if the input is integer.
        """
        filter_size = 3
        data2 = test_data.copy().astype(dtype)
        bkg = Background2D(data2, box_size, filter_size=filter_size)
        if data2.dtype.kind == 'f':
            assert bkg.background.dtype == data2.dtype
            assert bkg.background_rms.dtype == data2.dtype
            assert bkg.background_mesh.dtype == data2.dtype
            assert bkg.background_rms_mesh.dtype == data2.dtype
        else:
            assert np.issubdtype(bkg.background.dtype, np.floating)
            assert np.issubdtype(bkg.background_rms.dtype, np.floating)
            assert np.issubdtype(bkg.background_mesh.dtype, np.floating)
            assert np.issubdtype(bkg.background_rms_mesh.dtype,
                                 np.floating)
        assert bkg.n_pixels_map.dtype == int
        assert bkg.n_pixels_mesh.dtype == int
        assert_allclose(bkg.background, data2)
        assert_allclose(bkg.background_rms, bkg_rms)
        assert bkg.background_median == 1.0
        assert bkg.background_rms_median == 0.0
        assert bkg.n_pixels_map.shape == test_data.shape

    def test_background_nddata(self, test_data, bkg_rms, bkg_mesh,
                               bkg_rms_mesh, nddata_variant,
                               nddata_no_unit):
        """
        Test with NDData and CCDData, and also test units.
        """
        bkg = Background2D(nddata_variant, (25, 25), filter_size=3)
        assert isinstance(bkg.background, u.Quantity)
        assert isinstance(bkg.background_rms, u.Quantity)
        assert isinstance(bkg.background_median, u.Quantity)
        assert isinstance(bkg.background_rms_median, u.Quantity)

        bkg = Background2D(nddata_no_unit, (25, 25), filter_size=3)
        assert_allclose(bkg.background, test_data)
        assert_allclose(bkg.background_rms, bkg_rms)
        assert_allclose(bkg.background_mesh, bkg_mesh)
        assert_allclose(bkg.background_rms_mesh, bkg_rms_mesh)
        assert bkg.background_median == 1.0
        assert bkg.background_rms_median == 0.0

    def test_background_rect(self):
        """
        Regression test for interpolators with non-square input data.
        """
        data = np.arange(12).reshape(3, 4)
        rms = np.zeros((3, 4))
        bkg = Background2D(data, (1, 1), filter_size=1)
        assert_allclose(bkg.background, data, atol=0.005)
        assert_allclose(bkg.background_rms, rms)
        assert_allclose(bkg.background_mesh, data)
        assert_allclose(bkg.background_rms_mesh, rms)
        assert bkg.background_median == 5.5
        assert bkg.background_rms_median == 0.0

    def test_background_nonconstant_data(self, test_data, bkg_mesh):
        """
        Test on non-constant data to ensure that the background mesh
        is computed correctly and that the background is properly
        interpolated.
        """
        data = np.copy(test_data)
        data[25:50, 50:75] = 10.0
        bkg_low_res = np.copy(bkg_mesh)
        bkg_low_res[1, 2] = 10.0
        bkg1 = Background2D(data, (25, 25), filter_size=(1, 1))
        assert_allclose(bkg1.background_mesh, bkg_low_res)
        assert bkg1.background.shape == data.shape

        rng = np.random.default_rng(0)
        data = rng.normal(1.0, 0.1, (121, 289))
        mask = np.zeros(data.shape, dtype=bool)
        mask[50:100, 50:100] = True
        bkg = Background2D(data, (25, 25), mask=mask)
        assert np.mean(bkg.background) < 1.0
        assert np.mean(bkg.background_rms) < 1.0
        assert bkg.background_median < 1.0
        assert bkg.background_rms_median < 0.1
        assert bkg.n_pixels_mesh.shape == (5, 12)
        assert bkg.n_pixels_map.shape == data.shape

    def test_bkg_estimator_not_mutated(self, test_data):
        """
        Test that user-supplied estimator objects are not mutated.

        Background2D silences sigma clipping on the internal copy of the
        estimators. The original objects passed by the caller must be
        left unchanged.
        """
        sigclip = SigmaClip(sigma=3.0)
        bkg_est = MeanBackground(sigma_clip=sigclip)
        bkgrms_est = MedianBackground(sigma_clip=sigclip)

        # Remember the sigma_clip values before the call
        assert bkg_est.sigma_clip is sigclip
        assert bkgrms_est.sigma_clip is sigclip

        Background2D(test_data, (25, 25), bkg_estimator=bkg_est,
                     bkg_rms_estimator=bkgrms_est)

        # Check that original sigma_clip values are unchanged after the
        # call
        assert bkg_est.sigma_clip is sigclip
        assert bkgrms_est.sigma_clip is sigclip

    def test_filter_threshold_rms_mesh_before_mesh(self):
        """
        Test that accessing background_rms_mesh before background_mesh
        does not crash when filter_threshold is set.

        Background2D._bkg_stats is used by _selective_filter, which
        is called when filter_threshold is not None. It must still
        be available when background_mesh is computed even if
        background_rms_mesh was computed first.
        """
        data = np.ones((100, 100))
        data[25:50, 50:75] = 10.0
        bkg = Background2D(data, (25, 25), filter_size=(3, 3),
                           filter_threshold=9.0)

        # Access rms_mesh first, then the regular mesh
        rms_mesh = bkg.background_rms_mesh
        mesh = bkg.background_mesh
        assert rms_mesh.shape == (4, 4)
        assert mesh.shape == (4, 4)

        # Both should still give sensible results
        assert_allclose(mesh[1, 2], 1.0, atol=0.01)

    def test_rms_mesh_before_mesh_no_filter_threshold(self):
        """
        Test that accessing background_rms_mesh before background_mesh
        does not crash when filter_threshold is None (the default).

        _try_free_bkg_stats must not free _bkg_stats before
        background_mesh has been computed, otherwise _interpolate_grid
        receives None and raises a TypeError on np.isnan.
        """
        data = np.ones((101, 101))
        coverage_mask = np.zeros(data.shape, dtype=bool)
        coverage_mask[50:, 50:] = True
        bkg = Background2D(data, 50, coverage_mask=coverage_mask)

        # Access rms_mesh first, then the regular mesh
        rms_mesh = bkg.background_rms_mesh
        mesh = bkg.background_mesh
        assert rms_mesh.shape == mesh.shape

    def test_no_sigma_clipping(self, test_data):
        """
        Test bkg_estimator inputs without sigma clipping.
        """
        data = np.copy(test_data)
        data[10, 10] = 100.0
        bkg1 = Background2D(data, (25, 25), filter_size=(1, 1),
                            bkg_estimator=MeanBackground())
        bkg2 = Background2D(data, (25, 25), filter_size=(1, 1),
                            sigma_clip=None, bkg_estimator=MeanBackground())

        assert bkg2.background_mesh[0, 0] > bkg1.background_mesh[0, 0]

    def test_function_estimators(self, test_data):
        """
        Test with user-defined functions for bkg_estimator and
        bkg_rms_estimator.
        """
        def bkg_func(data, *, axis=None):
            return np.nanmean(data, axis=axis)

        def bkgrms_func(data, *, axis=None):
            return np.nanstd(data, axis=axis)

        bkg = Background2D(test_data, (25, 25), filter_size=(1, 1),
                           sigma_clip=None, bkg_estimator=bkg_func,
                           bkg_rms_estimator=bkgrms_func)
        assert_allclose(bkg.background, test_data)
        assert_allclose(bkg.background_rms, np.zeros(test_data.shape))

    def test_integer_input_background_not_truncated(self):
        """
        Test that the background is not truncated when the input data is
        integer type.
        """
        data = np.array([[1, 2], [1, 2]], dtype=int)
        bkg = Background2D(data, (2, 2), filter_size=(1, 1),
                           sigma_clip=None, bkg_estimator=MeanBackground())
        assert_allclose(bkg.background_mesh, [[1.5]])
        assert_allclose(bkg.background, np.full(data.shape, 1.5))
        assert np.issubdtype(bkg.background.dtype, np.floating)

    def test_resizing(self):
        """
        Test that the background mesh is resized correctly when the
        input data dimensions are not integer multiples of the box size.
        """
        shape1 = (128, 256)
        shape2 = (129, 256)
        box_size = (16, 16)
        data1 = np.ones(shape1)
        data2 = np.ones(shape2)
        bkg1 = Background2D(data1, box_size)
        bkg2 = Background2D(data2, box_size)
        assert bkg1.background_mesh.shape == (8, 16)
        assert bkg2.background_mesh.shape == (9, 16)
        assert bkg1.background.shape == shape1
        assert bkg2.background.shape == shape2

    @pytest.mark.parametrize('box_size', ([(25, 25), (23, 22)]))
    def test_background_mask(self, box_size, test_data, bkg_rms):
        """
        Test with an input mask with different box sizes.

        Note that box_size=(23, 22) tests the resizing of the image and
        mask.
        """
        data = np.copy(test_data)
        data[25:50, 25:50] = 100.0
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[25:50, 25:50] = True
        bkg = Background2D(data, box_size, filter_size=(1, 1), mask=mask,
                           bkg_estimator=MeanBackground())
        assert_allclose(bkg.background, test_data, rtol=2.0e-5)
        assert_allclose(bkg.background_rms, bkg_rms)

    def test_mask(self, test_data):
        """
        Test with an input mask.
        """
        data = np.copy(test_data)
        data[25:50, 25:50] = 100.0
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[25:50, 25:50] = True
        bkg1 = Background2D(data, (25, 25), filter_size=(1, 1), mask=None,
                            bkg_estimator=MeanBackground())
        assert np.all(bkg1.n_pixels_map == 625)
        assert np.all(bkg1.n_pixels_mesh == 625)
        assert bkg1.background.shape == data.shape
        assert_allclose(bkg1.background_mesh[0, 0], 1.0)
        assert_allclose(bkg1.background_mesh[1, 1], 100.0)
        assert np.all(bkg1.background_rms_mesh == 0.0)

        bkg2 = Background2D(data, (25, 25), filter_size=(1, 1), mask=mask,
                            bkg_estimator=MeanBackground())

        ngoodpix = test_data.size - 625
        assert np.count_nonzero(bkg2.n_pixels_map == 625) == ngoodpix
        assert np.count_nonzero(bkg2.n_pixels_mesh == 625) == 15
        assert bkg2.background.shape == data.shape
        assert_allclose(bkg2.background_mesh, 1.0)
        assert np.all(bkg2.background_rms_mesh == 0.0)

    @pytest.mark.parametrize('fill_value', [0.0, np.nan, -1.0])
    def test_coverage_mask(self, fill_value, test_data):
        """
        Test with an input coverage mask.
        """
        data = np.copy(test_data)
        data[:50, :50] = np.nan
        mask = np.isnan(data)

        bkg1 = Background2D(data, (25, 25), filter_size=(1, 1),
                            coverage_mask=mask, fill_value=fill_value,
                            bkg_estimator=MeanBackground())
        assert_equal(bkg1.background[:50, :50], fill_value)
        assert_equal(bkg1.background_rms[:50, :50], fill_value)

        # Test that combined mask and coverage_mask gives the same
        # results
        mask = np.zeros(test_data.shape, dtype=bool)
        coverage_mask = np.zeros(test_data.shape, dtype=bool)
        mask[:50, :25] = True
        coverage_mask[:50, 25:50] = True
        match = r'Input data contains non-finite \(NaN or infinity\) values'
        with pytest.warns(AstropyUserWarning, match=match):
            bkg2 = Background2D(data, (25, 25), filter_size=(1, 1), mask=mask,
                                coverage_mask=mask, fill_value=0.0,
                                bkg_estimator=MeanBackground())
        assert_allclose(bkg1.background_mesh, bkg2.background_mesh)
        assert_allclose(bkg1.background_rms_mesh, bkg2.background_rms_mesh)

    def test_mask_nonfinite(self, test_data):
        """
        Test that non-finite values in the input data are masked and a
        warning is issued.
        """
        data = test_data.copy()
        data[0, 0:50] = np.nan
        match = r'Input data contains non-finite \(NaN or infinity\) values'
        with pytest.warns(AstropyUserWarning, match=match):
            bkg = Background2D(data, (25, 25), filter_size=(1, 1))
        assert_allclose(bkg.background, test_data, rtol=1e-5)

    def test_mask_with_already_masked_nans(self, test_data):
        """
        Test masked invalid values.

        These tests should not issue a warning.
        """
        data = test_data.copy()
        data[50, 25:50] = np.nan
        mask = np.isnan(data)

        bkg = Background2D(data, (25, 25), filter_size=(1, 1), mask=mask)
        assert_allclose(bkg.background, test_data, rtol=1e-5)

        bkg = Background2D(data, (25, 25), filter_size=(1, 1),
                           coverage_mask=mask)
        assert bkg.background.shape == data.shape

        mask = np.zeros(data.shape, dtype=bool)
        coverage_mask = np.zeros(data.shape, dtype=bool)
        mask[50, 25:30] = True
        coverage_mask[50, 30:50] = True
        bkg = Background2D(data, (25, 25), filter_size=(1, 1), mask=mask,
                           coverage_mask=coverage_mask)
        assert bkg.background.shape == data.shape

    def test_masked_array(self, test_data):
        """
        Test that masked arrays are handled correctly.
        """
        data = test_data.copy()
        data[0, 0:50] = True
        mask = np.zeros(test_data.shape, dtype=bool)
        mask[0, 0:50] = True
        data_ma1 = np.ma.MaskedArray(test_data, mask=mask)
        data_ma2 = np.ma.MaskedArray(data, mask=mask)

        bkg1 = Background2D(data, (25, 25), filter_size=(1, 1))
        bkg2 = Background2D(data_ma1, (25, 25), filter_size=(1, 1))
        bkg3 = Background2D(data_ma2, (25, 25), filter_size=(1, 1))
        assert_allclose(bkg1.background, bkg2.background, rtol=1e-5)
        assert_allclose(bkg2.background, bkg3.background, rtol=1e-5)

    def test_completely_masked(self, test_data):
        """
        Test that an error is raised if all pixels are masked.
        """
        mask = np.ones(test_data.shape, dtype=bool)
        match = 'All input pixels are masked. Cannot compute a background.'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), mask=mask)
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), coverage_mask=mask)

        mask = np.zeros(test_data.shape, dtype=bool)
        coverage_mask = np.zeros(test_data.shape, dtype=bool)
        mask[:, 0:40] = True
        coverage_mask[:, 40:] = True
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), mask=mask,
                         coverage_mask=coverage_mask)

        data = test_data.copy()
        data[:] = np.nan
        match = r'Input data contains all non-finite \(NaN or infinity\)'
        with pytest.raises(ValueError, match=match):
            Background2D(data, (25, 25))

    def test_zero_padding(self, test_data, bkg_rms):
        """
        Test case where padding is added only on one axis.
        """
        bkg = Background2D(test_data, (25, 22), filter_size=(1, 1))
        assert_allclose(bkg.background, test_data, rtol=1e-5)
        assert_allclose(bkg.background_rms, bkg_rms)
        assert bkg.background_median == 1.0
        assert bkg.background_rms_median == 0.0

        bkg = Background2D(test_data, (22, 25), filter_size=(1, 1))
        assert_allclose(bkg.background, test_data, rtol=1e-5)
        assert_allclose(bkg.background_rms, bkg_rms)
        assert bkg.background_median == 1.0
        assert bkg.background_rms_median == 0.0

    def test_exclude_percentile(self, test_data):
        """
        Test that the exclude_percentile parameter excludes the correct
        pixels.
        """
        data = np.copy(test_data)
        data[0:50, 0:50] = np.nan
        match = r'Input data contains non-finite \(NaN or infinity\) values'
        with pytest.warns(AstropyUserWarning, match=match):
            bkg = Background2D(data, (25, 25), filter_size=(1, 1),
                               exclude_percentile=100.0)
        assert_equal(bkg.n_pixels_mesh[0:2, 0:2], np.zeros((2, 2)))
        assert bkg.n_pixels_mesh[-1, -1] == 625

        data = np.ones((111, 121))
        bkg = Background2D(data, box_size=10, exclude_percentile=100)
        assert_allclose(bkg.background_mesh, np.ones((12, 13)))

        data[:] = np.nan
        data[0, 0] = 1.0
        match1 = r'Input data contains non-finite \(NaN or infinity\) values'
        match2 = r'All boxes contain .* unmasked or finite pixels'
        ctx1 = pytest.warns(AstropyUserWarning, match=match1)
        ctx2 = pytest.raises(ValueError, match=match2)
        with ctx1, ctx2:
            Background2D(data, (10, 10))

    def test_filter_threshold(self, test_data, bkg_mesh):
        """
        Test that the filter_threshold parameter filters the correct
        pixels.
        """
        data = np.copy(test_data)
        data[25:50, 50:75] = 10.0
        bkg = Background2D(data, (25, 25), filter_size=(3, 3),
                           filter_threshold=9.0)
        assert_allclose(bkg.background, test_data)
        assert_allclose(bkg.background_mesh, bkg_mesh)
        bkg2 = Background2D(data, (25, 25), filter_size=(3, 3),
                            filter_threshold=11.0)  # No filtering
        assert bkg2.background_mesh[1, 2] == 10

    def test_filter_threshold_high(self, test_data, bkg_mesh):
        """
        Test that the filter_threshold parameter does not filter any
        pixels when it is set too high.
        """
        data = np.copy(test_data)
        data[25:50, 50:75] = 10.0
        ref_data = np.copy(bkg_mesh)
        ref_data[1, 2] = 10.0
        bkg = Background2D(data, (25, 25), filter_size=(3, 3),
                           filter_threshold=100.0)
        assert_allclose(bkg.background_mesh, ref_data)

    def test_filter_threshold_nofilter(self, test_data, bkg_mesh):
        """
        Test that the filter_threshold does not filter any pixels when
        the filter_size is (1, 1).
        """
        data = np.copy(test_data)
        data[25:50, 50:75] = 10.0
        ref_data = np.copy(bkg_mesh)
        ref_data[1, 2] = 10.0
        b = Background2D(data, (25, 25), filter_size=(1, 1),
                         filter_threshold=1.0)
        assert_allclose(b.background_mesh, ref_data)

    def test_scalar_sizes(self, test_data):
        """
        Test that scalar box_size and filter_size are correctly
        converted to tuples.
        """
        bkg1 = Background2D(test_data, (25, 25), filter_size=(3, 3))
        bkg2 = Background2D(test_data, 25, filter_size=3)
        assert_allclose(bkg1.background, bkg2.background)
        assert_allclose(bkg1.background_rms, bkg2.background_rms)

    def test_invalid_box_size(self, test_data):
        """
        Test that an error is raised if box_size has an invalid number
        of elements.
        """
        match = 'box_size must have 1 or 2 elements'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (5, 5, 3))

    def test_invalid_filter_size(self, test_data):
        """
        Test that an error is raised if filter_size has an invalid
        number of elements.
        """
        match = 'filter_size must have 1 or 2 elements'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (5, 5), filter_size=(3, 3, 3))

    def test_invalid_exclude_percentile(self, test_data):
        """
        Test that an error is raised if exclude_percentile is outside the
        range [0, 100].
        """
        match = 'exclude_percentile must be between 0 and 100'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (5, 5), exclude_percentile=-1)
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (5, 5), exclude_percentile=101)

    def test_mask_nomask(self, test_data):
        """
        Test that mask and coverage_mask can be set to np.ma.nomask and
        that the background is computed correctly.
        """
        bkg = Background2D(test_data, (25, 25), filter_size=(1, 1),
                           mask=np.ma.nomask)
        assert not bkg._has_mask

        bkg = Background2D(test_data, (25, 25), filter_size=(1, 1),
                           coverage_mask=np.ma.nomask)
        assert bkg.coverage_mask is None

    def test_invalid_mask(self, test_data):
        """
        Test that an error is raised if the mask has an invalid shape or
        number of dimensions.
        """
        match = 'data and mask must have the same shape'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), filter_size=(1, 1),
                         mask=np.zeros((2, 2)))

        match = 'mask must be a 2D array'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), filter_size=(1, 1),
                         mask=np.zeros((2, 2, 2)))

    def test_invalid_coverage_mask(self, test_data):
        """
        Test that an error is raised if the coverage_mask has an invalid
        shape or number of dimensions.
        """
        match = 'data and coverage_mask must have the same shape'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), filter_size=(1, 1),
                         coverage_mask=np.zeros((2, 2)))

        match = 'coverage_mask must be a 2D array'
        with pytest.raises(ValueError, match=match):
            Background2D(test_data, (25, 25), filter_size=(1, 1),
                         coverage_mask=np.zeros((2, 2, 2)))

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot_meshes(self, test_data):
        """
        Test the plot_meshes method.

        This test should run without any errors, but there is no return
        value.
        """
        bkg = Background2D(test_data, (25, 25))
        bkg.plot_meshes(outlines=True)

    def test_repr(self):
        """
        Test the __repr__ method.
        """
        data = np.ones((300, 500))
        bkg = Background2D(data, (74, 99))
        cls_repr = repr(bkg)
        assert cls_repr.startswith(f'{bkg.__class__.__name__}')

        mask = np.zeros(data.shape, dtype=bool)
        mask[0:10, 0:10] = True
        bkg = Background2D(data, (74, 99), mask=mask)
        cls_repr = repr(bkg)
        assert cls_repr.startswith(f'{bkg.__class__.__name__}')
        assert 'mask' in cls_repr

        bkg = Background2D(data, (74, 99), coverage_mask=mask)
        cls_repr = repr(bkg)
        assert cls_repr.startswith(f'{bkg.__class__.__name__}')
        assert 'coverage_mask' in cls_repr

    def test_str(self):
        """
        Test the __str__ method.
        """
        data = np.ones((300, 500))
        bkg = Background2D(data, (74, 99))
        cls_str = str(bkg)
        cls_name = bkg.__class__.__name__
        cls_name = f'{bkg.__class__.__module__}.{cls_name}'
        assert cls_str.startswith(f'<{cls_name}>')

    def test_masks(self):
        """
        Test that the input data is not modified when a mask is applied
        and that the same background is computed whether the non-finite
        values are masked or set to NaN.
        """
        arr = np.arange(25.0).reshape(5, 5)
        arr_orig = arr.copy()
        mask = np.zeros(arr.shape, dtype=bool)
        mask[0, 0] = np.nan
        mask[-1, 0] = np.nan
        mask[-1, -1] = np.nan
        mask[0, -1] = np.nan

        box_size = (2, 2)
        exclude_percentile = 100
        filter_size = 1
        bkg_estimator = MeanBackground()
        bkg1 = Background2D(arr, box_size, mask=mask,
                            exclude_percentile=exclude_percentile,
                            filter_size=filter_size,
                            bkg_estimator=bkg_estimator)
        bkgimg1 = bkg1.background
        assert_equal(arr, arr_orig)

        arr2 = arr.copy()
        arr2[mask] = np.nan
        arr3 = arr2.copy()
        match = r'Input data contains non-finite \(NaN or infinity\) values'
        with pytest.warns(AstropyUserWarning, match=match):
            bkg2 = Background2D(arr2, box_size, mask=None,
                                exclude_percentile=exclude_percentile,
                                filter_size=filter_size,
                                bkg_estimator=bkg_estimator)
        bkgimg2 = bkg2.background
        assert_equal(arr2, arr3)

        assert_allclose(bkgimg1, bkgimg2)

    @pytest.mark.parametrize('bkg_est', [MeanBackground(),
                                         SExtractorBackground()])
    def test_large_boxsize(self, bkg_est):
        """
        Test that when boxsize is the same as the image size that the
        input data is unchanged and that the background mesh is a single
        value equal to the background estimator applied to the entire
        image.
        """
        shape = (103, 107)
        data = np.ones(shape)
        data[50:55, 50:55] = 1000.0
        data[20:25, 20:25] = 1000.0
        box_size = data.shape
        filter_size = (3, 3)
        data_orig = data.copy()
        bkg = Background2D(data, box_size, filter_size=filter_size,
                           bkg_estimator=bkg_est)
        bkgim = bkg.background
        assert bkgim.shape == shape
        assert_equal(data, data_orig)

    def test_interpolator_keyword_deprecation(self, test_data):
        """
        Test that the interpolator keyword is deprecated.
        """
        match = 'BkgZoomInterpolator is deprecated'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            interp = BkgZoomInterpolator()

        match = "'interpolator' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bkg = Background2D(test_data, (25, 25), interpolator=interp)

        assert_allclose(bkg.background, test_data)

        bkg = Background2D(test_data, (25, 25))  # Should not raise
        assert_allclose(bkg.background, test_data)

    def test_background_box_size_one(self, test_data):
        """
        Test that when box_size is (1, 1) the background is equal to the
        input data.
        """
        bkg = Background2D(test_data, (1, 1), filter_size=(1, 1))
        assert_allclose(bkg.background, test_data, rtol=1e-5)

    def test_background_prime_dimensions(self):
        """
        Test with prime-number dimensions.
        """
        data = np.ones((97, 101))  # Prime dimensions
        bkg = Background2D(data, (10, 10))
        assert bkg.background.shape == data.shape
        assert_allclose(bkg.background, data, rtol=1e-5)

    def test_background_box_size_larger_than_image(self):
        """
        Test when box_size exceeds image dimensions.
        """
        data = np.ones((50, 60))
        bkg = Background2D(data, (100, 100))
        assert bkg.background.shape == data.shape
        # With box size larger than image, should get single mesh value
        assert bkg.background_mesh.shape == (1, 1)

        # Test with one box dimension larger than image
        bkg = Background2D(data, (100, 30))
        assert bkg.background.shape == data.shape
        assert bkg.background_mesh.shape == (1, 2)

        bkg = Background2D(data, (25, 100))
        assert bkg.background.shape == data.shape
        assert bkg.background_mesh.shape == (2, 1)

    def test_background_mesh_properties(self, test_data):
        """
        Test that the background mesh properties are consistent with the
        input data and box size.
        """
        bkg = Background2D(test_data, (25, 25))

        assert bkg.background_mesh.shape[0] * 25 >= test_data.shape[0]
        assert bkg.background_mesh.shape[1] * 25 >= test_data.shape[1]
        assert_allclose(bkg.background_median,
                        np.median(bkg.background_mesh))
        assert_allclose(bkg.background_rms_median,
                        np.median(bkg.background_rms_mesh))

    def test_input_data_not_mutated(self, test_data):
        """
        Test that the input data array is not modified by Background2D
        for various combinations of mask, coverage_mask, and box sizes
        that require padding.
        """
        # Basic case: no mask, no coverage_mask
        data = test_data.copy()
        data_orig = data.copy()
        Background2D(data, (25, 25))
        assert_equal(data, data_orig)

        # No mask, no coverage_mask, box_size same as image shape
        data = test_data.copy()
        data_orig = data.copy()
        Background2D(data, data.shape)
        assert_equal(data, data_orig)

        # With mask, box_size same as image shape
        data = test_data.copy()
        data_orig = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[10:20, 10:20] = True
        Background2D(data, data.shape, mask=mask)
        assert_equal(data, data_orig)

        # With coverage_mask, box_size same as image shape
        data = test_data.copy()
        data_orig = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[10:20, 10:20] = True
        Background2D(data, data.shape, coverage_mask=mask)
        assert_equal(data, data_orig)

        # With outliers in the data (exercises sigma-clipping path)
        data = test_data.copy()
        data[10, 10] = 1000.0
        data_orig = data.copy()
        Background2D(data, (25, 25))
        assert_equal(data, data_orig)

        # With mask
        data = test_data.copy()
        data_orig = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[10:20, 10:20] = True
        Background2D(data, (25, 25), mask=mask)
        assert_equal(data, data_orig)

        # With coverage_mask
        data = test_data.copy()
        data_orig = data.copy()
        coverage_mask = np.zeros(data.shape, dtype=bool)
        coverage_mask[50:, 50:] = True
        Background2D(data, (25, 25), coverage_mask=coverage_mask)
        assert_equal(data, data_orig)

        # With both mask and coverage_mask
        data = test_data.copy()
        data_orig = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[10:20, 10:20] = True
        coverage_mask = np.zeros(data.shape, dtype=bool)
        coverage_mask[50:, 50:] = True
        Background2D(data, (25, 25), mask=mask, coverage_mask=coverage_mask)
        assert_equal(data, data_orig)

        # With box size that requires padding (not an integer multiple)
        data = test_data.copy()
        data_orig = data.copy()
        Background2D(data, (23, 22))  # 100 / 23 and 100 / 22 are not integers
        assert_equal(data, data_orig)

        # Padding with mask
        data = test_data.copy()
        data_orig = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[5:15, 5:15] = True
        Background2D(data, (23, 22), mask=mask)
        assert_equal(data, data_orig)

        # Padding with coverage_mask
        data = test_data.copy()
        data_orig = data.copy()
        coverage_mask = np.zeros(data.shape, dtype=bool)
        coverage_mask[60:, 60:] = True
        Background2D(data, (23, 22), coverage_mask=coverage_mask)
        assert_equal(data, data_orig)

    def test_input_masked_array_not_mutated(self, test_data):
        """
        Test that a masked-array input is not modified by Background2D.
        """
        data_values = test_data.copy()
        mask = np.zeros(data_values.shape, dtype=bool)
        mask[10:20, 10:20] = True
        data_ma = np.ma.MaskedArray(data_values, mask=mask)
        data_values_orig = data_values.copy()
        mask_orig = mask.copy()

        Background2D(data_ma, (25, 25))

        assert_equal(data_ma.data, data_values_orig)
        assert_equal(data_ma.mask, mask_orig)

        # Box size requiring padding
        data_ma2 = np.ma.MaskedArray(data_values.copy(), mask=mask.copy())
        Background2D(data_ma2, (23, 22))
        assert_equal(data_ma2.data, data_values_orig)
        assert_equal(data_ma2.mask, mask_orig)


def test_deprecations(test_data):
    data = test_data.copy()
    bkg = Background2D(data, (25, 25))
    match = 'was deprecated'
    with pytest.warns(AstropyDeprecationWarning, match=match):
        assert bkg.npixels_mesh.shape == (4, 4)
    with pytest.warns(AstropyDeprecationWarning, match=match):
        assert bkg.npixels_map.shape == data.shape



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.morphology import data_properties, gini


class TestDataPropertiesPositionalKwargs:
    """
    Test data_properties warns for positional optional args.
    """

    def setup_method(self):
        self.data = np.random.default_rng(0).random((10, 10))

    def test_positional_warns(self):
        mask = np.zeros((10, 10), dtype=bool)
        match = 'data_properties'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            data_properties(self.data, mask)

    def test_keyword_no_warning(self):
        mask = np.zeros((10, 10), dtype=bool)
        data_properties(self.data, mask=mask)


class TestGiniPositionalKwargs:
    """
    Test gini warns for positional optional args.
    """

    def test_positional_warns(self):
        data = np.arange(100, dtype=float)
        mask = np.zeros(100, dtype=bool)
        match = 'gini'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            gini(data, mask)

    def test_keyword_no_warning(self):
        data = np.arange(100, dtype=float)
        mask = np.zeros(100, dtype=bool)
        gini(data, mask=mask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the examples module.
"""

from numpy.testing import assert_allclose

from photutils.datasets import make_4gaussians_image, make_100gaussians_image


def test_make_4gaussians_image():
    """
    Test the make_4gaussians_image function.
    """
    shape = (100, 200)
    data_sum = 177189.58
    image = make_4gaussians_image()
    assert image.shape == shape
    assert_allclose(image.sum(), data_sum, rtol=1.0e-6)


def test_make_4gaussians_image_no_noise():
    """
    Test the make_4gaussians_image function with no noise.
    """
    shape = (100, 200)
    image = make_4gaussians_image(noise=False)
    assert image.shape == shape
    assert image.min() >= 0


def test_make_100gaussians_image():
    """
    Test the make_100gaussians_image function.
    """
    shape = (300, 500)
    data_sum = 826059.53
    image = make_100gaussians_image()
    assert image.shape == shape
    assert_allclose(image.sum(), data_sum, rtol=1.0e-6)


def test_make_100gaussians_image_no_noise():
    """
    Test the make_100gaussians_image function with no noise.
    """
    shape = (300, 500)
    image = make_100gaussians_image(noise=False)
    assert image.shape == shape
    assert image.min() >= 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing tools for generating kernels for matching point
spread functions.
"""

from .fourier import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
from .windows import *  # noqa: F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the detect module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.stats import SigmaClip
from numpy.testing import assert_allclose, assert_equal

from photutils.segmentation.detect import detect_sources, detect_threshold
from photutils.segmentation.utils import make_2dgaussian_kernel
from photutils.utils.exceptions import NoDetectionsWarning

DATA = np.array([[0, 1, 0], [0, 2, 0], [0, 0, 0]]).astype(float)
REF1 = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]])


class TestDetectThreshold:
    def test_nsigma(self):
        """
        Test basic n_sigma.
        """
        threshold = detect_threshold(DATA, n_sigma=0.1)
        ref = 0.4 * np.ones((3, 3))
        assert_allclose(threshold, ref)

        threshold = detect_threshold(DATA << u.uJy, n_sigma=0.1)
        assert isinstance(threshold, u.Quantity)
        assert_allclose(threshold.value, ref)

    def test_nsigma_zero(self):
        """
        Test n_sigma=0.
        """
        threshold = detect_threshold(DATA, n_sigma=0.0)
        ref = (1.0 / 3.0) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background(self):
        """
        Test background.
        """
        threshold = detect_threshold(DATA, n_sigma=1.0, background=1)
        ref = (5.0 / 3.0) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_background_image(self):
        """
        Test background image.
        """
        background = np.ones((3, 3))
        threshold = detect_threshold(DATA, n_sigma=1.0, background=background)
        ref = (5.0 / 3.0) * np.ones((3, 3))
        assert_allclose(threshold, ref)

        threshold = detect_threshold(DATA << u.Jy, n_sigma=1.0,
                                     background=background << u.Jy)
        assert isinstance(threshold, u.Quantity)
        assert_allclose(threshold.value, ref)

    def test_background_badshape(self):
        """
        Test background badshape.
        """
        wrong_shape = np.zeros((2, 2))
        match = 'input background is 2D, then it must have the same shape'
        with pytest.raises(ValueError, match=match):
            detect_threshold(DATA, n_sigma=2.0, background=wrong_shape)

    def test_error(self):
        """
        Test error.
        """
        threshold = detect_threshold(DATA, n_sigma=1.0, error=1)
        ref = (4.0 / 3.0) * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_error_image(self):
        """
        Test error image.
        """
        error = np.ones((3, 3))
        threshold = detect_threshold(DATA, n_sigma=1.0, error=error)
        ref = (4.0 / 3.0) * np.ones((3, 3))
        assert_allclose(threshold, ref)

        threshold = detect_threshold(DATA << u.Jy, n_sigma=1.0,
                                     error=error << u.Jy)
        assert isinstance(threshold, u.Quantity)
        assert_allclose(threshold.value, ref)

    def test_error_badshape(self):
        """
        Test error badshape.
        """
        wrong_shape = np.zeros((2, 2))
        match = 'If input error is 2D, then it must have the same shape'
        with pytest.raises(ValueError, match=match):
            detect_threshold(DATA, n_sigma=2.0, error=wrong_shape)

    def test_background_error(self):
        """
        Test background error.
        """
        threshold = detect_threshold(DATA, n_sigma=2.0, background=10.0,
                                     error=1.0)
        ref = 12.0 * np.ones((3, 3))
        assert_allclose(threshold, ref)

        threshold = detect_threshold(DATA << u.Jy, n_sigma=2.0,
                                     background=10.0 * u.Jy,
                                     error=1.0 * u.Jy)
        assert isinstance(threshold, u.Quantity)
        assert_allclose(threshold.value, ref)

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            detect_threshold(DATA << u.Jy, n_sigma=2.0, background=10.0,
                             error=1.0 * u.Jy)
        with pytest.raises(ValueError, match=match):
            detect_threshold(DATA << u.Jy, n_sigma=2.0, background=10.0 * u.m,
                             error=1.0 * u.Jy)

    def test_background_error_images(self):
        """
        Test background error images.
        """
        background = np.ones((3, 3)) * 10.0
        error = np.ones((3, 3))
        threshold = detect_threshold(DATA, n_sigma=2.0, background=background,
                                     error=error)
        ref = 12.0 * np.ones((3, 3))
        assert_allclose(threshold, ref)

    def test_image_mask(self):
        """
        Test detection with image_mask.

        Set sigma=10 and iters=1 to prevent sigma clipping after
        applying the mask.
        """
        mask = REF1.astype(bool)
        sigma_clip = SigmaClip(sigma=10, maxiters=1)
        threshold = detect_threshold(DATA, n_sigma=1.0, error=0, mask=mask,
                                     sigma_clip=sigma_clip)
        ref = (1.0 / 8.0) * np.ones((3, 3))
        assert_equal(threshold, ref)

    def test_invalid_sigma_clip(self):
        """
        Test invalid sigma clip.
        """
        match = 'sigma_clip must be a SigmaClip object'
        with pytest.raises(TypeError, match=match):
            detect_threshold(DATA, 1.0, sigma_clip=10)


class TestDetectSources:
    def setup_class(self):
        self.data = np.array([[0, 1, 0], [0, 2, 0],
                              [0, 0, 0]]).astype(float)
        self.refdata = np.array([[0, 1, 0], [0, 1, 0], [0, 0, 0]])

        kernel = make_2dgaussian_kernel(2.0, size=3)
        self.kernel = kernel

    def test_detection(self):
        """
        Test basic detection.
        """
        segm = detect_sources(self.data, threshold=0.9, n_pixels=2)
        assert_equal(segm.data, self.refdata)

        assert segm.data.dtype == np.int32
        assert segm.labels.dtype == np.int32

        segm = detect_sources(self.data << u.uJy, threshold=0.9 * u.uJy,
                              n_pixels=2)
        assert_equal(segm.data, self.refdata)

        match = 'must all have the same units'
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data << u.uJy, threshold=0.9, n_pixels=2)
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data, threshold=0.9 * u.Jy, n_pixels=2)
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data << u.uJy, threshold=0.9 * u.m, n_pixels=2)

    def test_small_sources(self):
        """
        Test detection where sources are smaller than n_pixels size.
        """
        match = 'No sources were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            detect_sources(self.data, threshold=0.9, n_pixels=5)

    def test_n_pixels(self):
        """
        Test removal of sources whose size is less than n_pixels.

        Regression tests for #663.
        """
        data = np.zeros((8, 8))
        data[0:4, 0] = 1
        data[0, 0:4] = 1
        data[3, 3:] = 2
        data[3:, 3] = 2

        segm = detect_sources(data, 0, n_pixels=4)
        assert segm.n_labels == 2
        assert segm.data.dtype == np.int32

        # Removal of labels with size less than n_pixels
        # dtype should still be np.int32
        segm = detect_sources(data, 0, n_pixels=8)
        assert segm.n_labels == 1
        assert segm.data.dtype == np.int32

        segm = detect_sources(data, 0, n_pixels=9)
        assert segm.n_labels == 1
        assert segm.data.dtype == np.int32

        data = np.zeros((8, 8))
        data[0:4, 0] = 1
        data[0, 0:4] = 1
        data[3, 2:] = 2
        data[3:, 2] = 2
        data[5:, 3] = 2

        n_pixels = np.arange(9, 14)
        for n_pixels in np.arange(9, 14):
            segm = detect_sources(data, 0, n_pixels=n_pixels)
            assert segm.n_labels == 1
            assert segm.areas[0] == 13

        match = 'No sources were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            detect_sources(data, 0, n_pixels=14)

    def test_zerothresh(self):
        """
        Test detection with zero threshold.
        """
        segm = detect_sources(self.data, threshold=0.0, n_pixels=2)
        assert_equal(segm.data, self.refdata)

    def test_zerodet(self):
        """
        Test detection with large threshold giving no detections.
        """
        match = 'No sources were found'
        with pytest.warns(NoDetectionsWarning, match=match):
            detect_sources(self.data, threshold=7, n_pixels=2)

    def test_8connectivity(self):
        """
        Test detection with connectivity=8.
        """
        data = np.eye(3)
        segm = detect_sources(data, threshold=0.9, n_pixels=1, connectivity=8)
        assert_equal(segm.data, data)

    def test_4connectivity(self):
        """
        Test detection with connectivity=4.
        """
        data = np.eye(3)
        ref = np.diag([1, 2, 3])
        segm = detect_sources(data, threshold=0.9, n_pixels=1, connectivity=4)
        assert_equal(segm.data, ref)

    def test_n_pixels_nonint(self):
        """
        Test if an error is raised when npixel is noninteger.
        """
        match = 'n_pixels must be a positive integer'
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data, threshold=1, n_pixels=0.1)

    def test_n_pixels_negative(self):
        """
        Test if an error is raised when npixel is negative.
        """
        match = 'n_pixels must be a positive integer'
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data, threshold=1, n_pixels=-1)

    @pytest.mark.parametrize('connectivity', [0, -1, 3, 6, 10])
    def test_connectivity_invalid(self, connectivity):
        """
        Test if an error is raised when connectivity is invalid.
        """
        match = f'Invalid connectivity={connectivity}'
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data, threshold=1, n_pixels=1,
                           connectivity=connectivity)

    def test_mask(self):
        """
        Test mask.
        """
        data = np.zeros((11, 11))
        data[3:8, 3:8] = 5.0
        mask = np.zeros(data.shape, dtype=bool)
        mask[4:6, 4:6] = True
        segm1 = detect_sources(data, 1.0, 1.0)
        segm2 = detect_sources(data, 1.0, 1.0, mask=mask)
        assert segm2.areas[0] == segm1.areas[0] - mask.sum()

        # Mask with all True
        mask = np.ones(data.shape, dtype=bool)
        match = 'mask must not be True for every pixel'
        with pytest.raises(ValueError, match=match):
            detect_sources(data, 1.0, 1.0, mask=mask)

    def test_mask_shape(self):
        """
        Test mask shape.
        """
        match = 'mask must have the same shape as the input image'
        with pytest.raises(ValueError, match=match):
            detect_sources(self.data, 1.0, 1.0, mask=np.ones((5, 5)))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _utils module.
"""

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_array_equal

from photutils.centroids._utils import (_gaussian2d_moments,
                                        _process_data_mask, _validate_data,
                                        _validate_gaussian_inputs,
                                        _validate_mask_shape)


class TestValidateData:
    """
    Tests for _validate_data.
    """

    def test_converts_to_float(self):
        data = np.array([[1, 2], [3, 4]], dtype=int)
        result = _validate_data(data)
        assert result.dtype == float

    def test_2d_default(self):
        data = np.ones((4, 4))
        result = _validate_data(data)
        assert result.shape == (4, 4)

    def test_wrong_ndim_raises(self):
        data = np.ones((4, 4, 4))
        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            _validate_data(data, ndim=2)

    def test_1d_valid(self):
        data = np.ones(10)
        result = _validate_data(data, ndim=1)
        assert result.ndim == 1

    def test_ndim_none_accepts_any_shape(self):
        for shape in [(5,), (5, 5), (5, 5, 5)]:
            result = _validate_data(np.ones(shape), ndim=None)
            assert result.shape == shape


class TestValidateMaskShape:
    """
    Tests for _validate_mask_shape.
    """

    def test_none_mask_passes(self):
        data = np.ones((4, 4))
        _validate_mask_shape(data, None)

    def test_matching_shape_passes(self):
        data = np.ones((4, 4))
        mask = np.zeros((4, 4), dtype=bool)
        _validate_mask_shape(data, mask)

    def test_mismatched_shape_raises(self):
        data = np.ones((4, 4))
        mask = np.zeros((2, 2), dtype=bool)
        match = 'data and mask must have the same shape'
        with pytest.raises(ValueError, match=match):
            _validate_mask_shape(data, mask)


class TestProcessDataMask:
    """
    Tests for _process_data_mask.
    """

    def test_finite_data_unchanged(self):
        data = np.array([[1.0, 2.0], [3.0, 4.0]])
        result = _process_data_mask(data, None)
        assert_array_equal(result, data)

    def test_nan_fills_and_warns(self):
        data = np.array([[1.0, np.nan], [3.0, 4.0]])
        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            result = _process_data_mask(data, None)
        assert np.isnan(result[0, 1])

    def test_nan_fills_with_custom_fill_value(self):
        data = np.array([[1.0, np.nan], [3.0, 4.0]])
        with pytest.warns(AstropyUserWarning):
            result = _process_data_mask(data, None, fill_value=0.0)
        assert result[0, 1] == 0.0

    def test_mask_fills_fill_value(self):
        data = np.array([[1.0, 2.0], [3.0, 4.0]])
        mask = np.array([[False, True], [False, False]])
        result = _process_data_mask(data, mask, fill_value=0.0)
        assert result[0, 1] == 0.0

    def test_masked_nan_no_warning(self):
        data = np.array([[1.0, np.nan], [3.0, 4.0]])
        mask = np.array([[False, True], [False, False]])
        # Masked NaN should not trigger a warning
        result = _process_data_mask(data, mask, fill_value=0.0)
        assert result[0, 1] == 0.0

    def test_ndim_validation(self):
        data = np.ones((4, 4, 4))
        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            _process_data_mask(data, None, ndim=2)

    def test_mask_shape_validation(self):
        data = np.ones((4, 4))
        mask = np.zeros((2, 2), dtype=bool)
        match = 'data and mask must have the same shape'
        with pytest.raises(ValueError, match=match):
            _process_data_mask(data, mask)

    def test_masked_array_no_mutation(self):
        """
        Input mask must not be mutated when data is a MaskedArray.
        """
        masked_data = np.ma.array([[1.0, 2.0], [3.0, 4.0]],
                                  mask=[[False, True], [False, False]])
        input_mask = np.zeros((2, 2), dtype=bool)
        input_mask_orig = input_mask.copy()

        _process_data_mask(masked_data, input_mask)
        assert_array_equal(input_mask, input_mask_orig)

    def test_masked_array_returns_ndarray(self):
        """
        MaskedArray input must return a plain ndarray, not MaskedArray.
        """
        masked_data = np.ma.array([[1.0, 2.0], [3.0, 4.0]],
                                  mask=[[False, True], [False, False]])
        result = _process_data_mask(masked_data, None)
        assert not isinstance(result, np.ma.MaskedArray)
        assert isinstance(result, np.ndarray)

    def test_masked_array_mask_combined(self):
        """
        MaskedArray mask and keyword mask are combined correctly.
        """
        masked_data = np.ma.array([[1.0, 2.0], [3.0, 4.0]],
                                  mask=[[False, True], [False, False]])
        extra_mask = np.array([[True, False], [False, False]])
        result = _process_data_mask(masked_data, extra_mask, fill_value=0.0)
        # Both (0,0) from extra_mask and (0,1) from MaskedArray should
        # be 0
        assert result[0, 0] == 0.0
        assert result[0, 1] == 0.0


class TestValidateGaussianInputs:
    """
    Tests for _validate_gaussian_inputs.
    """

    @pytest.fixture
    def gauss_data(self):
        rng = np.random.default_rng(0)
        return np.abs(rng.standard_normal((20, 20))) + 1.0

    def test_no_error_no_mask(self, gauss_data):
        data, mask, error = _validate_gaussian_inputs(gauss_data, None, None)
        assert error is None
        assert data.shape == gauss_data.shape
        assert mask.shape == gauss_data.shape
        assert mask.dtype == bool

    def test_error_shape_mismatch_raises(self, gauss_data):
        error = np.ones((5, 5))
        match = 'data and error must have the same shape'
        with pytest.raises(ValueError, match=match):
            _validate_gaussian_inputs(gauss_data, None, error)

    def test_nan_in_error_sets_combined_mask(self, gauss_data):
        error = np.ones_like(gauss_data)
        error[5, 5] = np.nan
        data, combined_mask, _ = _validate_gaussian_inputs(
            gauss_data, None, error)
        assert combined_mask[5, 5]
        assert data[5, 5] == 0.0

    def test_nan_in_data_sets_combined_mask(self, gauss_data):
        gauss_data = gauss_data.copy()
        gauss_data[3, 3] = np.nan
        match = 'Input data contains non-finite values'
        with pytest.warns(AstropyUserWarning, match=match):
            data, combined_mask, _ = _validate_gaussian_inputs(
                gauss_data, None, None)
        assert combined_mask[3, 3]
        assert data[3, 3] == 0.0

    def test_input_mask_not_mutated(self, gauss_data):
        error = np.ones_like(gauss_data)
        error[5, 5] = np.nan
        mask = np.zeros(gauss_data.shape, dtype=bool)
        mask_orig = mask.copy()
        _validate_gaussian_inputs(gauss_data, mask, error)
        assert_array_equal(mask, mask_orig)

    def test_input_error_not_mutated(self, gauss_data):
        error = np.ones_like(gauss_data)
        error[5, 5] = np.nan
        error_orig = error.copy()
        _validate_gaussian_inputs(gauss_data, None, error)
        assert_array_equal(error, error_orig)

    def test_data_not_mutated_when_error_nan(self, gauss_data):
        """
        data must not be mutated when NaN in error extends combined_mask
        beyond the positions already handled by _process_data_mask.

        When mask=None and data is clean, _process_data_mask returns
        the original object without copying. If error then contributes
        new NaN positions, the code must not write 0.0 directly into
        the caller's array.
        """
        error = np.ones_like(gauss_data)
        error[5, 5] = np.nan
        data_orig = gauss_data.copy()
        _validate_gaussian_inputs(gauss_data, None, error)
        assert_array_equal(gauss_data, data_orig)

    def test_error_zeroed_at_combined_mask(self, gauss_data):
        mask = np.zeros(gauss_data.shape, dtype=bool)
        mask[2, 2] = True
        error = np.ones_like(gauss_data)
        error[5, 5] = np.nan
        _, _, out_error = _validate_gaussian_inputs(gauss_data, mask, error)
        assert out_error[2, 2] == 0.0   # masked by input mask
        assert out_error[5, 5] == 0.0   # masked by NaN in error

    def test_no_copy_when_no_invalids(self, gauss_data):
        """
        Test that error values are unchanged when there are no NaNs in
        data or error, and mask is None.
        """
        error = np.ones_like(gauss_data)
        _, _, out_error = _validate_gaussian_inputs(gauss_data, None, error)
        assert_array_equal(out_error, error)


class TestGaussian2DMoments:
    """
    Tests for _gaussian2d_moments.
    """

    def test_symmetric_gaussian(self):
        """
        Circular Gaussian: centroid and equal stddevs are recovered.
        """
        xcen, ycen, std = 25.0, 25.0, 5.0
        model = Gaussian2D(1.0, xcen, ycen, x_stddev=std, y_stddev=std)
        y, x = np.mgrid[0:50, 0:50]
        data = model(x, y)
        (amplitude,
         x_mean,
         y_mean,
         x_stddev,
         y_stddev,
         theta,
         ) = _gaussian2d_moments(data)

        assert_allclose(amplitude, 1.0, atol=0.01)
        assert_allclose(x_mean, xcen, atol=0.01)
        assert_allclose(y_mean, ycen, atol=0.01)
        assert_allclose(x_stddev, std, atol=0.1)
        assert_allclose(y_stddev, std, atol=0.1)
        assert_allclose(theta, 0, atol=0.05)

    @pytest.mark.parametrize('theta_in',
                             np.deg2rad((0, 22, 37, 45, 60, 88, 90)))
    def test_asymmetric_gaussian_theta(self, theta_in):
        """
        Axis-aligned elliptical Gaussian (theta=0): centroid and axis
        stddevs are recovered. The larger sigma maps to x_stddev.
        """
        ampl = 3.5
        xcen, ycen = 30.0, 20.0
        x_std, y_std = 6.0, 3.0
        model = Gaussian2D(ampl, xcen, ycen, x_stddev=x_std, y_stddev=y_std,
                           theta=theta_in)
        y, x = np.mgrid[0:50, 0:50]
        data = model(x, y)
        (amplitude,
         x_mean,
         y_mean,
         x_stddev,
         y_stddev,
         theta,
         ) = _gaussian2d_moments(data)

        assert_allclose(amplitude, ampl, atol=0.01)
        assert_allclose(x_mean, xcen, atol=0.02)
        assert_allclose(y_mean, ycen, atol=0.02)
        assert_allclose(x_stddev, x_std, atol=0.1)
        assert_allclose(y_stddev, y_std, atol=0.1)
        assert_allclose(theta, theta_in, atol=0.05)
